import 'package:endlich/scheduler.dart';
import 'package:flutter/widgets.dart';

abstract class StateMachine<State> with ChangeNotifier {
  StateMachine() {
    createInitialState();
  }

  static final Scheduler _scheduler = Scheduler()..initialize();

  StateConfig<State>? _config;

  late State state;

  bool _mounted = true;
  bool get mounted => _mounted;

  /// A development-only way to access [state] .
  ///
  /// This is useful for tests.
  State get debugState {
    late State result;

    assert(() {
      result = state;
      return true;
    }(), '');
    return result;
  }

  void createInitialState();

  @override
  void dispose() {
    _config?.dispose();
    _config = null;
    _mounted = false;
    super.dispose();
  }

  void transition(State Function(StateConfig<State> config) cb) {
    _config?.dispose();

    _scheduler.schedule(() {
      final config = StateConfig<State>();
      _config = config;
      config.configure(cb);

      /// check wheter there was an immediate transition
      if (config.isActive) {
        state = config.state;

        for (var enter in config._onEnter) {
          if (!config.isActive) break;
          enter();
        }

        if (config.isActive) notifyListeners();
      }
    });
  }
}

class StateConfig<State> {
  StateConfig();

  final List<void Function()> _onEnter = [];
  final List<void Function()> _onLeave = [];

  late final State state;
  late bool _debugInConfigure;
  bool _isActive = true;
  bool get isActive => _isActive;

  void dispose() {
    _isActive = false;
    for (var leave in _onLeave) {
      leave();
    }
  }

  void configure(State Function(StateConfig<State> config) cb) {
    _debugInConfigure = true;
    state = cb(this);
    _debugInConfigure = false;
  }

  void onEnter(void Function() cb) {
    if (!_debugInConfigure || !_isActive) {
      throw StateError('''
It is no longer possible to add an onEnter callback in state "${state.runtimeType}".
Consider to use config.isActive before.''');
    }

    _onEnter.add(cb);
  }

  void onLeave(void Function() cb) {
    if (!_isActive) {
      throw StateError('''
It is no longer possible to add an onLeave callback in state "$runtimeType".
Consider using config.isActive before.''');
    }
    _onLeave.add(cb);
  }
}
