import 'package:flutter_test/flutter_test.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/mockito.dart';

import 'package:endlich/src/machine.dart';

part 'machine_test.freezed.dart';

void main() {
  test('initial state', () {
    final machine = TestMachine();
    expect(machine.debugState, isA<_Foo>());
  });

  test('transition', () {
    final machine = TestMachine();

    final foo = machine.debugState as _Foo;
    foo.toBar();
    expect(machine.debugState, isA<_Bar>());

    (machine.debugState as _Bar).next();
    expect(machine.debugState, isA<_Baz>());
  });

  test('transition in body of state creation', () {
    final machine = TestMachine(
      onFoo: (foo, config, machine) {
        machine.toBar(0);
      },
      onBar: (bar, config, machine) {
        machine.toBaz(0);
      },
    );

    expect(machine.debugState, isA<_Baz>());
  });

  group('StateConfig', () {
    group('isActive', () {
      test('before and after transition', () {
        var beforeTransition = false;
        var afterTransition = false;
        TestMachine(
          onFoo: (foo, config, machine) {
            beforeTransition = config.isActive;
            machine.toBar(0);
            afterTransition = config.isActive;
          },
        );
        expect(beforeTransition, isTrue);
        expect(afterTransition, isFalse);
      });

      test('disposing machine will result in false', () {
        late StateConfig<Test> config;
        final m = TestMachine(
          onFoo: (foo, c, machine) {
            config = c;
          },
        );
        m.dispose();
        expect(config.isActive, isFalse);
      });
    });

    group('onEnter', () {
      test('callback execution in right order', () {
        final fooEnter = OnEnterCb();
        final fooEnter2 = OnEnterCb();

        final barEnter = OnEnterCb();
        final barEnter2 = OnEnterCb();
        final machine = TestMachine(
          onFoo: (foo, config, machine) {
            config.onEnter(fooEnter);
            config.onEnter(fooEnter2);
          },
          onBar: (bar, config, machine) {
            config.onEnter(barEnter);
            config.onEnter(barEnter2);
          },
        );

        (machine.debugState as _Foo).toBar();

        verifyInOrder([fooEnter(), fooEnter2(), barEnter(), barEnter2()]);
        verifyNoMoreInteractions(fooEnter);
        verifyNoMoreInteractions(fooEnter2);
        verifyNoMoreInteractions(barEnter);
        verifyNoMoreInteractions(barEnter2);
      });

      test('adding a callback after state has become inactive will throw', () {
        late StateConfig<Test> config;

        TestMachine(onFoo: (foo, c, machine) {
          config = c;
          machine.toBar(0);
        });

        expect(
            () => config.onEnter(() {}),
            throwsA(isA<StateError>()
                .having((p0) => p0.message, 'assertion message', contains('''
It is no longer possible to add an onEnter callback in state "_\$_Foo".
Consider to use config.isActive before.'''))));
      });

      test('adding a callback after dispose will throw', () {
        late StateConfig<Test> config;

        final m = TestMachine(onFoo: (foo, c, machine) {
          config = c;
        });
        m.dispose();

        expect(
            () => config.onEnter(() {}),
            throwsA(isA<StateError>()
                .having((p0) => p0.message, 'assertion message', contains('''
It is no longer possible to add an onEnter callback in state "_\$_Foo".
Consider to use config.isActive before.'''))));
      });

      test(
          'leaving the state in body of state creation will not execute any callbacks added before',
          () {
        final fooEnter = OnEnterCb();
        TestMachine(onFoo: (foo, c, machine) {
          c.onEnter(fooEnter);
          machine.toBar(0);
        });
        verifyZeroInteractions(fooEnter);
      });

      test(
          'leaving the state in body of callback  will prevent further execution of subsequent callbacks',
          () {
        final fooEnter = OnEnterCb();

        final fooEnter2 = OnEnterCb();
        TestMachine(onFoo: (foo, c, machine) {
          c.onEnter(fooEnter);
          c.onEnter(() {
            machine.toBar(0);
          });
          c.onEnter(fooEnter2);
        });
        verifyZeroInteractions(fooEnter2);
        verify(fooEnter()).called(1);
        verifyNoMoreInteractions(fooEnter);
      });
    });

    group('onLeave', () {
      test('execute callbacks in right order', () {
        final fooLeave = OnLeaveCb();
        final fooLeave2 = OnLeaveCb();
        final fooLeave3 = OnLeaveCb();
        final machine = TestMachine(
          onFoo: (foo, config, machine) {
            config.onLeave(fooLeave);
            config.onEnter(() {
              config.onLeave(fooLeave3);
            });
            config.onLeave(fooLeave2);
          },
        );

        verifyZeroInteractions(fooLeave);
        verifyZeroInteractions(fooLeave2);
        verifyZeroInteractions(fooLeave3);
        (machine.debugState as _Foo).toBar();
        verifyInOrder([
          fooLeave(),
          fooLeave2(),
          fooLeave3(),
        ]);
        verifyNoMoreInteractions(fooLeave);
        verifyNoMoreInteractions(fooLeave2);
        verifyNoMoreInteractions(fooLeave3);
      });

      test('it is forbidden to add callbackas after state has been left', () {
        late StateConfig<Test> config;
        final m = TestMachine(
          onFoo: (foo, c, machine) {
            config = c;
          },
        );
        (m.debugState as _Foo).toBar();
        expect(
            () => config.onLeave(() {}),
            throwsA(isA<StateError>()
                .having((p0) => p0.message, 'assertion message', contains('''
It is no longer possible to add an onLeave callback in state "StateConfig<Test>".
Consider using config.isActive before.'''))));
      });
    });

    group('mixed onEnter and onLeave:', () {
      test('execute onLeave before onEnter on transition', () {
        final fooEnter = OnLeaveCb();
        final barEnter = OnLeaveCb();
        final fooLeave = OnLeaveCb();
        final machine = TestMachine(
          onFoo: (foo, config, machine) {
            config.onEnter(fooEnter);
            config.onLeave(fooLeave);
          },
          onBar: (bar, config, machine) {
            config.onEnter(barEnter);
          },
        );
        (machine.debugState as _Foo).toBar();
        verifyInOrder([
          fooEnter(),
          fooLeave(),
          barEnter(),
        ]);
      });
    });
  });

  group('machine interaction', () {
    group('nested', () {
      test('leave/enter actions are executed in order', () {
        final rootFooEnter = OnEnterCb1();
        final m1BarEnter = OnEnterCb2();
        final m2BarEnter = OnEnterCb3();
        final rootFooLeave = OnLeaveCb1();
        final m1BarLeave = OnLeaveCb2();
        final m2BarLeave = OnLeaveCb3();

        final machine2 = TestMachine(
          onBar: (bar, config, machine) {
            config.onEnter(m2BarEnter);
            config.onLeave(m2BarLeave);
          },
        );

        final machine1 = TestMachine(
          onBar: (bar, config, machine) {
            config.onEnter(() => machine2.toBar(0));
            config.onEnter(m1BarEnter);

            config.onLeave(() => machine2.toFoo(0));
            config.onLeave(m1BarLeave);
          },
        );

        final rootMachine = TestMachine(
          onFoo: (foo, config, machine) {
            config.onEnter(() => machine1.toBar(0));
            config.onEnter(rootFooEnter);

            config.onLeave(() => machine1.toFoo(0));
            config.onLeave(rootFooLeave);
          },
        );

        (rootMachine.debugState as _Foo).toBar();

        verifyInOrder([
          rootFooEnter(),
          m1BarEnter(),
          m2BarEnter(),
          m2BarLeave(),
          m1BarLeave(),
          rootFooLeave(),
        ]);
      });
    });

    test('listen to parent machine', () {
      final mRootBarEnter = OnEnterCb1();
      final m1barEnter = OnEnterCb2();
      final m2barEnter = OnEnterCb3();
      final mRootFooLeave = OnLeaveCb1();
      final m1fooLeave = OnLeaveCb2();
      final m2fooLeave = OnLeaveCb3();

      final rootMachine = TestMachine(
        onFoo: (foo, config, machine) {
          config.onLeave(mRootFooLeave);
        },
        onBar: (bar, config, machine) {
          config.onEnter(mRootBarEnter);
        },
      );
      final machine1 = TestMachine(
        onFoo: (foo, config, machine) {
          listener() {
            machine.toBar(0);
          }

          config.onEnter(() {
            rootMachine.addListener(listener);
            config.onLeave(() {
              rootMachine.removeListener(listener);
            });
          });

          config.onLeave(m1fooLeave);
        },
        onBar: (bar, config, machine) {
          config.onEnter(m1barEnter);
        },
      );
      final machine2 = TestMachine(
        onFoo: (foo, config, machine) {
          listener() {
            machine.toBar(0);
          }

          config.onEnter(() {
            machine1.addListener(listener);
            config.onLeave(() {
              machine1.removeListener(listener);
            });
          });
          config.onLeave(m2fooLeave);
        },
        onBar: (bar, config, machine) {
          config.onEnter(() {
            m2barEnter();
          });
        },
      );
      (rootMachine.debugState as _Foo).toBar();
      verifyInOrder([
        mRootFooLeave(),
        mRootBarEnter(),
        m1fooLeave(),
        m1barEnter(),
        m2fooLeave(),
        m2barEnter(),
      ]);

      verifyNoMoreInteractions(m1barEnter);
      verifyNoMoreInteractions(m2barEnter);
    });
  });
}

class TestMachine extends StateMachine<Test> {
  TestMachine({
    this.onFoo,
    this.onBar,
    this.onBaz,
  });

  void Function(_Foo foo, StateConfig<Test> config, TestMachine machine)? onFoo;
  void Function(_Bar bar, StateConfig<Test> config, TestMachine machine)? onBar;
  void Function(_Baz baz, StateConfig<Test> config, TestMachine machine)? onBaz;

  void toFoo(int i) => transition((config) {
        final s = _Foo(
          i: i,
          toBar: () => toBar(i + 1),
          next: () => toBar(i + 1),
        );
        onFoo?.call(s, config, this);
        return s;
      });

  void toBar(int i) => transition((config) {
        final s = _Bar(
            i: i,
            toFoo: () => toFoo(i + 1),
            toBaz: () => toBaz(i + 1),
            next: () => toBaz(i + 1));
        onBar?.call(s, config, this);
        return s;
      });

  void toBaz(int i) => transition((config) {
        final s = _Baz(
            i: i,
            toFoo: () => toFoo(i + 1),
            toBar: () => toBar(i + 1),
            next: () => toFoo(i + 1));
        onBaz?.call(s, config, this);
        return s;
      });

  @override
  void createInitialState() => toFoo(0);
}

@freezed
class Test with _$Test {
  factory Test.foo({
    required int i,
    required void Function() toBar,
    required void Function() next,
  }) = _Foo;
  factory Test.bar({
    required int i,
    required void Function() toFoo,
    required void Function() toBaz,
    required void Function() next,
  }) = _Bar;

  factory Test.baz({
    required int i,
    required void Function() toFoo,
    required void Function() toBar,
    required void Function() next,
  }) = _Baz;
}

class OnEnterCb extends Mock {
  void call();
}

class OnLeaveCb extends Mock {
  void call();
}

class OnEnterCb1 extends Mock {
  void call();
}

class OnEnterCb2 extends Mock {
  void call();
}

class OnEnterCb3 extends Mock {
  void call();
}

class OnEnterCb4 extends Mock {
  void call();
}

class OnLeaveCb1 extends Mock {
  void call();
}

class OnLeaveCb2 extends Mock {
  void call();
}

class OnLeaveCb3 extends Mock {
  void call();
}

class OnLeaveCb4 extends Mock {
  void call();
}
